#include "ltc.h"

/*

SST Task Frame structure

    +------------+
 0  | frame lock |  -> SST task frame lock
    +------------+
 1  |  cond obj  |  -> condition object
    +------------+
 2  |  func obj  |  -> function object body
    +------------+
 3  |    pc1     |  -> stealable / future expression pc
    +------------+
 4  |    pc2     |  -> condition  expression pc
    +------------+
 5  |   result   |  -> result 
    +------------+
 6  |  status    |  -> speculative/mandatory/not-needed
    |            |  -> stealable/future | executing/determined
    +------------+
 7  |    env     |  -> environment-list
    +------------+


thr_data Frame

    +-------------+
 1  | LOCK_STACK  |  -> lock
    +-------------+
 2  | PT_2_PAREN  |  -> pointer to parent process
    +-------------+
 3  | CHILD_LIST  |  -> pointer to child linked-list
    +-------------+
 4  | PT_2_FRAME  |  -> pointer to stolen frame
    +-------------+
 5  |  THR_STATE  |  -> Thread state : suspend or run
    +-------------+
 6  | PT_BAS_MAN  |  -> SST Mandatory base pointer
    +-------------+
 7  | PT_HED_MAN  |  -> SST Mandatory head pointer
    +-------------+
 8  | PT_TAL_MAN  |  -> SST Mandatory tail pointer
    +-------------+
 9  | KILLED_RES  |  -> Killed result
    +-------------+
10  | PT_BAS_SPE  |  -> SST Speculative base pointer
    +-------------+
11  | PT_HED_SPE  |  -> SST Speculative head pointer
    +-------------+
12  | PT_TAL_SPE  |  -> SST Speculative tail pointer
    +-------------+
13  | THR_STATUS  |  -> thread status : PRIMARY, SUBSIDIARY_M, SUBSIDIARY_S
    +-------------+
14  | THREAD_ID   |  -> thread id
    +-------------+
15  | THR_KILLED  |  -> should this task be killed ?
    +-------------+
16  | COND_EXEC   |  -> condition execution flag
    +-------------+
17  |  PT_2_HEAD  |  -> pointer to free list's head 
    +-------------+
18  |  PT_2_TAIL  |  -> pointer to free list's tail
    +-------------+
19  |  PT_BSEV_NC |  -> pointer to base of environment stack ( No Cond )
    +-------------+
20  |  PT_TPEV_NC |  -> pointer to top of environment stack ( No Cond )
    +-------------+
21  |  PT_BSEV_CO |  -> pointer to base of environment stack ( Cond )
    +-------------+
22  |  PT_TPEV_CO |  -> pointer to base of environment stack ( Cond )
    +-------------+



*/

#define BOTTOM_SZ       100

#define LOCK_STACK        1   
#define PT_2_PAREN        2   
#define CHILD_LIST        3   
#define PT_2_FRAME        4   
#define THR_STATE         5   
#define PT_BAS_MAN        6   
#define PT_HED_MAN        7
#define PT_TAL_MAN        8
#define KILLED_RES        9
#define PT_BAS_SPE       10
#define PT_HED_SPE       11
#define PT_TAL_SPE       12
#define THR_STATUS       13
#define THREAD_ID        14
#define THR_KILLED       15
#define COND_EXEC        16
#define PT_2_HEAD        17
#define PT_2_TAIL        18
#define PT_BSEV_NC       19
#define PT_TPEV_NC       20
#define PT_BSEV_CO       21
#define PT_TPEV_CO       22
#define PT_ENV_CO        23

#define PROMOTEDP         1

#define NOT_NEEDED        4
#define PAR_CONSTR        8
/* #define   FUTURE      16        not used */
#define EXECUTING        32
#define DETERMINED       64       /* for STEALABLE */

#define MAN_MAN_NO_COND   1
#define SUB_MAN_COND      2
#define SUB_SUB_NO_COND   3
#define SUB_SUB_COND      4
#define MAN_SUB_MAN       5

#define SPECULATIVE_SST   1
#define MANDATORY_SST     2

#define PRIMARY         0xA
#define SUBSIDIARY_M    0xB
#define SUBSIDIARY_S    0xC

#define YES             0xD
#define NO              0xE

#define RUN               3
#define SUSPEND           4
#define SLEEP             5

#define KILL_BY_PAR       1
#define KILL_BY_CON       2

#define thr_data ((unsigned int *)(stack_bottom+28))

#define   STEALABLE_CONST 8
#define   FUTURE_CONST    0
#define init_sstfr_status(sstfr, constr) ((sstfr)->status = (constr))

/* future:    par_constr() => FUTURE_CONST */
/* stealable: par_constr() => STEALABLE_CONST */
#define par_constr(sstfr) ((sstfr)->status & PAR_CONSTR)

/* future:    sstfr_status() => EXECUTING or NOT_STOLEN */
/* stealable: sstfr_status() => DETERMINED, EXECUTING, or NOT_STOLEN */
#define NOT_STOLEN     0
#define sstfr_status(sstfr) (((sstfr)->status & DETERMINED) \
                              ? DETERMINED : (sstfr)->status & EXECUTING)
#define set_sstfr_status(sstfr, stat) ((sstfr)->status |= (stat))

#define promoted_sstfr_p(sstfr) ((sstfr)->status & PROMOTEDP)
#define set_promote_sstfr(sstfr) ((sstfr)->status |= PROMOTEDP)
#define killed_sstfr_p(sstfr) ((sstfr)->status & NOT_NEEDED)
#define set_kill_sstfr(sstfr) ((sstfr)->status |= NOT_NEEDED)

#ifdef SHE_ENV
# define NO_COND 0
# define COND    1
#endif

typedef struct child_list{
   unsigned int * child;
   struct child_list *next;
} child_list_t;


typedef struct {
  int          active;
  condition_t  wait_cond;
  mutex_t      lock;
} thread_info_t;

extern int            iretaddrx_she_seq_fut();
extern int            iretaddrx_she_seq_st();
extern sst_frame_t  * real_alloc_sst_frame();
extern void           init_sst_frame_free_list();
extern sst_frame_t  * alloc_sst_frame();
extern void           free_sst_frame(sst_frame_t *);

extern void           cut_connections_with(child_list_t *,
                                           unsigned int *);
extern void           kill_my_descendant();
extern void           kill_my_descendant_with_value(unsigned int);

/* from here */

extern int            iretaddrx_con_she_seq_fut(void);
extern int            iretaddrx_con_she_seq_st(void);
extern int            steal_sst(void);

extern void           init_she_scheduler(int);
extern void           init_she_subthread(int);
extern void           shestack_alloc(thread_data_t **, int);
extern void           clean_up_sst(void);
extern void           clean_up_thr_data(void);
extern void           iret_she_steal(void);
extern void           she_steal(void);
extern void           push_sst_frame(sst_frame_t *, int);
extern void           free_child_list(child_list_t *);
extern void           start_bci_pri(int);
extern void           start_bci_pri_aux(int);
extern void           start_bci_sub(int);
extern int            set_condition(void);

extern child_list_t * add_child_list(child_list_t *, unsigned int *);
extern sst_frame_t  * pop_sst_frame(int);
extern sst_frame_t  * steal_sst_frame(void);
extern sst_frame_t  * steal_sst_frame_aux(int, int);
extern int print_stat(thread_data_t **, int);
extern thread_data_t ** pthrs;
extern thread_data_t ** sthrs;

#ifdef SHE_ENV
void push_env(unsigned int, int);
void copy_env_to_cstack(sst_frame_t *);
void pop_env_to_cstack(sst_frame_t *);
void copy_env_to_con_stack(sst_frame_t *);
#endif

/* #define SST_STACK_SIZE 0x1000 */
#define SST_SIZE 0x20000
#define ENV_STACK_SIZE (SST_SIZE*3)
